@@ -17,17 +17,7 @@ impl Search {
     invocation_directory: &Path,
   ) -> SearchResult<Self> {
     match search_config {
-      SearchConfig::FromInvocationDirectory => {
-        let justfile = Self::justfile(&invocation_directory)?;
-
-        let working_directory = Self::working_directory_from_justfile(&justfile)?;
-
-        Ok(Self {
-          justfile,
-          working_directory,
-        })
-      }
-
+      SearchConfig::FromInvocationDirectory => Self::find_next(invocation_directory),
       SearchConfig::FromSearchDirectory { search_directory } => {
         let search_directory = Self::clean(invocation_directory, search_directory);
 
@@ -40,7 +30,6 @@ impl Search {
           working_directory,
         })
       }
-
       SearchConfig::WithJustfile { justfile } => {
         let justfile = Self::clean(invocation_directory, justfile);
 
@@ -51,7 +40,6 @@ impl Search {
           working_directory,
         })
       }
-
       SearchConfig::WithJustfileAndWorkingDirectory {
         justfile,
         working_directory,
@@ -62,13 +50,24 @@ impl Search {
     }
   }
 
+  pub(crate) fn find_next(starting_dir: &Path) -> SearchResult<Self> {
+    let justfile = Self::justfile(starting_dir)?;
+
+    let working_directory = Self::working_directory_from_justfile(&justfile)?;
+
+    Ok(Self {
+      justfile,
+      working_directory,
+    })
+  }
+
   pub(crate) fn init(
     search_config: &SearchConfig,
     invocation_directory: &Path,
   ) -> SearchResult<Self> {
     match search_config {
       SearchConfig::FromInvocationDirectory => {
-        let working_directory = Self::project_root(&invocation_directory)?;
+        let working_directory = Self::project_root(invocation_directory)?;
 
         let justfile = working_directory.join(DEFAULT_JUSTFILE_NAME);
 
@@ -174,7 +173,7 @@ impl Search {
           io_error,
           directory: directory.to_owned(),
         })?;
-        for project_root_child in PROJECT_ROOT_CHILDREN.iter().cloned() {
+        for project_root_child in PROJECT_ROOT_CHILDREN.iter().copied() {
           if entry.file_name() == project_root_child {
             return Ok(directory.to_owned());
           }
